#include "ehs/system/CPU.h"
#include "ehs/Log.h"
#include "ehs/io/File.h"
#include "ehs/json/Json.h"
#include "ehs/system/Thread.h"

namespace ehs
{
    #ifdef EHS_OS_LINUX
		UInt_64 CPU::TSC_Freq = 0;
    #endif

    const bool CPU::hasFPU = RetrieveFPU();
    const bool CPU::hasVME = RetrieveVME();
    const bool CPU::hasDE = RetrieveDE();
    const bool CPU::hasPSE = RetrievePSE();
    const bool CPU::hasTSC = RetrieveTSC();
    const bool CPU::hasMSR = RetrieveMSR();
    const bool CPU::hasPAE = RetrievePAE();
    const bool CPU::hasMCE = RetrieveMCE();
    const bool CPU::hasCX8 = RetrieveCX8();
    const bool CPU::hasAPIC = RetrieveAPIC();
    const bool CPU::hasSEP = RetrieveSEP();
    const bool CPU::hasMTRR = RetrieveMTRR();
    const bool CPU::hasPGE = RetrievePGE();
    const bool CPU::hasMCA = RetrieveMCA();
    const bool CPU::hasCMOV = RetrieveCMOV();
    const bool CPU::hasPSE_36 = RetrievePSE_36();
    const bool CPU::hasPSN = RetrievePSN();
    const bool CPU::hasCLFSH = RetrieveCLFSH();
    const bool CPU::hasDS = RetrieveDS();
    const bool CPU::hasACPI = RetrieveACPI();
    const bool CPU::hasMMX = RetrieveMMX();
    const bool CPU::hasFXSR = RetrieveFXSR();
    const bool CPU::hasSSE = RetrieveSSE();
    const bool CPU::hasSSE2 = RetrieveSSE2();
    const bool CPU::hasSS = RetrieveSS();
    const bool CPU::hasHTT = RetrieveHTT();
    const bool CPU::hasTM = RetrieveTM();
    const bool CPU::hasIA64 = RetrieveIA64();
    const bool CPU::hasPBE = RetrievePBE();
    const bool CPU::hasSSE3 = RetrieveSSE3();
    const bool CPU::hasPCLMULQDQ = RetrievePCLMULQDQ();
    const bool CPU::hasDTES64 = RetrieveDTES64();
    const bool CPU::hasMONITOR = RetrieveMONITOR();
    const bool CPU::hasVMX = RetrieveVMX();
    const bool CPU::hasSMX = RetrieveSMX();
    const bool CPU::hasEST = RetrieveEST();
    const bool CPU::hasTM2 = RetrieveTM2();
    const bool CPU::hasSSSE3 = RetrieveSSSE3();
    const bool CPU::hasCNXT_ID = RetrieveCNXT_ID();
    const bool CPU::hasSDBG = RetrieveSDBG();
    const bool CPU::hasFMA = RetrieveFMA();
    const bool CPU::hasCX16 = RetrieveCX16();
    const bool CPU::hasXTPR = RetrieveXTPR();
    const bool CPU::hasPDCM = RetrievePDCM();
    const bool CPU::hasPCID = RetrievePCID();
    const bool CPU::hasDCA = RetrieveDCA();
    const bool CPU::hasSSE4_1 = RetrieveSSE4_1();
    const bool CPU::hasSSE4_2 = RetrieveSSE4_2();
    const bool CPU::hasX2APIC = RetrieveX2APIC();
    const bool CPU::hasMOVBE = RetrieveMOVBE();
    const bool CPU::hasPOPCNT = RetrievePOPCNT();
    const bool CPU::hasTSC_DEADLINE = RetrieveTSC_DEADLINE();
    const bool CPU::hasAES = RetrieveAES();
    const bool CPU::hasXSAVE = RetrieveXSAVE();
    const bool CPU::hasOSXSAVE = RetrieveOSXSAVE();
    const bool CPU::hasAVX = RetrieveAVX();
    const bool CPU::hasF16C = RetrieveF16C();
    const bool CPU::hasRDRND = RetrieveRDRND();
    const bool CPU::hasHYPERVISOR = RetrieveHYPERVISOR();
    const bool CPU::hasAVX2 = RetrieveAVX2();
    const bool CPU::hasAVX512F = RetrieveAVX512F();
    const bool CPU::hasRDSEED = RetrieveRDSEED();
    const bool CPU::hasADX = RetrieveADX();

	Architecture CPU::GetArchitecture()
	{
		#if defined(EHS_ARCH_X64)
			return Architecture::X64;
		#elif defined(EHS_ARCH_ARM64)
			return Architecture::ARM64;
		#else
			return Architecture::UNKNOWN;
		#endif
	}

	UInt_8 CPU::PointerSize()
	{
		return sizeof(void*);
	}

    Endianness CPU::GetEndianness()
    {
        #if defined(EHS_LITTLE_ENDIAN)
            return Endianness::LE;
        #elif defined(EHS_BIG_ENDIAN)
            return Endianness::BE;
        #else
            UInt_16 tmp = 1;
            if (((Byte*)&tmp)[0] == 1)
                return Endianness::LE;

            return Endianness::BE;
        #endif
    }

	UInt_64 CPU::GetTSC_Freq()
	{
		#if defined(EHS_OS_WINDOWS)
			LARGE_INTEGER frequency = {};
			QueryPerformanceFrequency(&frequency);

			return frequency.QuadPart;
        #elif defined(EHS_OS_LINUX)
            if (!TSC_Freq)
                TSC_Freq = RetrieveTSC_Freq();

            return TSC_Freq;
		#endif

		return 0;
	}

	UInt_64 CPU::GetTSC()
	{
		#if defined(EHS_OS_WINDOWS)
			LARGE_INTEGER count = {};
            QueryPerformanceCounter(&count);

			return count.QuadPart;
        #elif defined(EHS_OS_LINUX)
            TSC tsc;
            RDTSCP(&tsc);

            #if defined(EHS_ARCH_X64)
				UInt_64 result = 0;

                #if defined(EHS_LITTLE_ENDIAN)
                    ((UInt_32*)&result)[0] = tsc.lowCount;
                    ((UInt_32*)&result)[1] = tsc.highCount;
                #elif defined(EHS_BIG_ENDIAN)
                    ((UInt_32*)&result)[0] = tsc.highCount;
                    ((UInt_32*)&result)[1] = tsc.lowCount;
                #endif

                return result;
            #elif defined(EHS_ARCH_X86)
                return tsc.lowPart;
            #endif
		#endif

		return 0;
	}

    UInt_8 CPU::GetSteppingId()
    {
        return (UInt_8)GetInfoBits() & 0x00001111;
    }

    UInt_8 CPU::GetModelId()
    {
        return (UInt_8)GetInfoBits() & 0x11110000;
    }

    UInt_8 CPU::GetFamilyId()
    {
        return (UInt_8)(GetInfoBits() >> 8) & 0x00001111;
    }

    UInt_8 CPU::GetProcessorTypeId()
    {
        return (UInt_8)(GetInfoBits() >> 12) & 0x00000011;
    }

    UInt_8 CPU::GetExtModelId()
    {
        return (UInt_8)(GetInfoBits() >> 16) & 0x00001111;
    }

    UInt_8 CPU::GetExtFamilyId()
    {
        return (UInt_8)(GetInfoBits() >> 20);
    }

    bool CPU::RetrieveFPU()
    {
        #ifdef EHS_ARCH_X64
            return GetFeatureBits_1() & 0b00000000000000000000000000000001;
        #else
	        return false;
        #endif
    }

    bool CPU::RetrieveVME()
    {
        #ifdef EHS_ARCH_X64
            return GetFeatureBits_1() & 0b00000000000000000000000000000010;
        #else
	        return false;
        #endif
    }

    bool CPU::RetrieveDE()
    {
        #ifdef EHS_ARCH_X64
            return GetFeatureBits_1() & 0b00000000000000000000000000000100;
        #else
	        return false;
        #endif
    }

    bool CPU::RetrievePSE()
    {
        #ifdef EHS_ARCH_X64
            return GetFeatureBits_1() & 0b00000000000000000000000000001000;
        #else
	        return false;
        #endif
    }

    bool CPU::RetrieveTSC()
    {
        #ifdef EHS_ARCH_X64
            return GetFeatureBits_1() & 0b00000000000000000000000000010000;
        #else
	            return false;
        #endif
    }

    bool CPU::RetrieveMSR()
    {
        #ifdef EHS_ARCH_X64
            return GetFeatureBits_1() & 0b00000000000000000000000000100000;
        #else
	        return false;
        #endif
    }

    bool CPU::RetrievePAE()
    {
        #ifdef EHS_ARCH_X64
            return GetFeatureBits_1() & 0b00000000000000000000000001000000;
        #else
	        return false;
        #endif
    }

    bool CPU::RetrieveMCE()
    {
        #ifdef EHS_ARCH_X64
            return GetFeatureBits_1() & 0b00000000000000000000000010000000;
        #else
	        return false;
        #endif
    }

    bool CPU::RetrieveCX8()
    {
        #ifdef EHS_ARCH_X64
            return GetFeatureBits_1() & 0b00000000000000000000000100000000;
        #else
	        return false;
        #endif
    }

    bool CPU::RetrieveAPIC()
    {
        #ifdef EHS_ARCH_X64
            return GetFeatureBits_1() & 0b00000000000000000000001000000000;
        #else
	        return false;
        #endif
    }

    bool CPU::RetrieveSEP()
    {
        #ifdef EHS_ARCH_X64
            return GetFeatureBits_1() & 0b00000000000000000000100000000000;
        #else
	        return false;
        #endif
    }

    bool CPU::RetrieveMTRR()
    {
        #ifdef EHS_ARCH_X64
            return GetFeatureBits_1() & 0b00000000000000000001000000000000;
        #else
	        return false;
        #endif
    }

    bool CPU::RetrievePGE()
    {
        #ifdef EHS_ARCH_X64
            return GetFeatureBits_1() & 0b00000000000000000010000000000000;
        #else
	        return false;
        #endif
    }

    bool CPU::RetrieveMCA()
    {
        #ifdef EHS_ARCH_X64
            return GetFeatureBits_1() & 0b00000000000000000100000000000000;
        #else
	        return false;
        #endif
    }

    bool CPU::RetrieveCMOV()
    {
        #ifdef EHS_ARCH_X64
            return GetFeatureBits_1() & 0b00000000000000001000000000000000;
        #else
	        return false;
        #endif
    }

    bool CPU::RetrievePAT()
    {
        #ifdef EHS_ARCH_X64
            return GetFeatureBits_1() & 0b00000000000000010000000000000000;
        #else
	        return false;
        #endif
    }

    bool CPU::RetrievePSE_36()
    {
        #ifdef EHS_ARCH_X64
            return GetFeatureBits_1() & 0b00000000000000100000000000000000;
        #else
	        return false;
        #endif
    }

    bool CPU::RetrievePSN()
    {
        #ifdef EHS_ARCH_X64
            return GetFeatureBits_1() & 0b00000000000001000000000000000000;
        #else
	        return false;
        #endif
    }

    bool CPU::RetrieveCLFSH()
    {
        #ifdef EHS_ARCH_X64
            return GetFeatureBits_1() & 0b00000000000010000000000000000000;
        #else
	        return false;
        #endif
    }

    bool CPU::RetrieveDS()
    {
        #ifdef EHS_ARCH_X64
            return GetFeatureBits_1() & 0b00000000001000000000000000000000;
        #else
	        return false;
        #endif
    }

    bool CPU::RetrieveACPI()
    {
        #ifdef EHS_ARCH_X64
            return GetFeatureBits_1() & 0b00000000010000000000000000000000;
        #else
	        return false;
        #endif
    }

    bool CPU::RetrieveMMX()
    {
        #ifdef EHS_ARCH_X64
            return GetFeatureBits_1() & 0b00000000100000000000000000000000;
        #else
	            return false;
        #endif
    }

    bool CPU::RetrieveFXSR()
    {
        #ifdef EHS_ARCH_X64
            return GetFeatureBits_1() & 0b00000001000000000000000000000000;
        #else
	            return false;
        #endif
    }

    bool CPU::RetrieveSSE()
    {
        #ifdef EHS_ARCH_X64
            return GetFeatureBits_1() & 0b00000010000000000000000000000000;
        #else
	        return false;
        #endif
    }

    bool CPU::RetrieveSSE2()
    {
        #ifdef EHS_ARCH_X64
            return GetFeatureBits_1() & 0b00000100000000000000000000000000;
        #else
	        return false;
        #endif
    }

    bool CPU::RetrieveSS()
    {
        #ifdef EHS_ARCH_X64
            return GetFeatureBits_1() & 0b00001000000000000000000000000000;
        #else
	        return false;
        #endif
    }

    bool CPU::RetrieveHTT()
    {
        #ifdef EHS_ARCH_X64
            return GetFeatureBits_1() & 0b00010000000000000000000000000000;
        #else
	        return false;
        #endif
    }

    bool CPU::RetrieveTM()
    {
        #ifdef EHS_ARCH_X64
            return GetFeatureBits_1() & 0b00100000000000000000000000000000;
        #else
	        return false;
        #endif
    }

    bool CPU::RetrieveIA64()
    {
        #ifdef EHS_ARCH_X64
            return GetFeatureBits_1() & 0b01000000000000000000000000000000;
        #else
	        return false;
        #endif
    }

    bool CPU::RetrievePBE()
    {
        #ifdef EHS_ARCH_X64
            return GetFeatureBits_1() & 0b10000000000000000000000000000000;
        #else
	        return false;
        #endif
    }

    bool CPU::RetrieveSSE3()
    {
        #ifdef EHS_ARCH_X64
            return GetFeatureBits_2() & 0b00000000000000000000000000000001;
        #else
	        return false;
        #endif
    }

    bool CPU::RetrievePCLMULQDQ()
    {
        #ifdef EHS_ARCH_X64
            return GetFeatureBits_2() & 0b00000000000000000000000000000010;
        #else
	        return false;
        #endif
    }

    bool CPU::RetrieveDTES64()
    {
        #ifdef EHS_ARCH_X64
            return GetFeatureBits_2() & 0b00000000000000000000000000000100;
        #else
	        return false;
        #endif
    }

    bool CPU::RetrieveMONITOR()
    {
        #ifdef EHS_ARCH_X64
            return GetFeatureBits_2() & 0b00000000000000000000000000001000;
        #else
	        return false;
        #endif
    }

    bool CPU::RetrieveDS_CPL()
    {
        #ifdef EHS_ARCH_X64
            return GetFeatureBits_2() & 0b00000000000000000000000000010000;
        #else
	        return false;
        #endif
    }

    bool CPU::RetrieveVMX()
    {
        #ifdef EHS_ARCH_X64
            return GetFeatureBits_2() & 0b00000000000000000000000000100000;
        #else
	        return false;
        #endif
    }

    bool CPU::RetrieveSMX()
    {
        #ifdef EHS_ARCH_X64
            return GetFeatureBits_2() & 0b00000000000000000000000001000000;
        #else
	        return false;
        #endif
    }

    bool CPU::RetrieveEST()
    {
        #ifdef EHS_ARCH_X64
            return GetFeatureBits_2() & 0b00000000000000000000000010000000;
        #else
	        return false;
        #endif
    }

    bool CPU::RetrieveTM2()
    {
        #ifdef EHS_ARCH_X64
            return GetFeatureBits_2() & 0b00000000000000000000000100000000;
        #else
	        return false;
        #endif
    }

    bool CPU::RetrieveSSSE3()
    {
        #ifdef EHS_ARCH_X64
            return GetFeatureBits_2() & 0b00000000000000000000001000000000;
        #else
	        return false;
        #endif
    }

    bool CPU::RetrieveCNXT_ID()
    {
        #ifdef EHS_ARCH_X64
            return GetFeatureBits_2() & 0b00000000000000000000010000000000;
        #else
	        return false;
        #endif
    }

    bool CPU::RetrieveSDBG()
    {
        #ifdef EHS_ARCH_X64
            return GetFeatureBits_2() & 0b00000000000000000000100000000000;
        #else
	        return false;
        #endif
    }

    bool CPU::RetrieveFMA()
    {
        #ifdef EHS_ARCH_X64
            return GetFeatureBits_2() & 0b00000000000000000001000000000000;
        #else
	        return false;
        #endif
    }

    bool CPU::RetrieveCX16()
    {
        #ifdef EHS_ARCH_X64
            return GetFeatureBits_2() & 0b00000000000000000010000000000000;
        #else
	        return false;
        #endif
    }

    bool CPU::RetrieveXTPR()
    {
        #ifdef EHS_ARCH_X64
            return GetFeatureBits_2() & 0b00000000000000000100000000000000;
        #else
	        return false;
        #endif
    }

    bool CPU::RetrievePDCM()
    {
        #ifdef EHS_ARCH_X64
            return GetFeatureBits_2() & 0b00000000000000001000000000000000;
        #else
	        return false;
        #endif
    }

    bool CPU::RetrievePCID()
    {
        #ifdef EHS_ARCH_X64
            return GetFeatureBits_2() & 0b00000000000000100000000000000000;
        #else
	        return false;
        #endif
    }

    bool CPU::RetrieveDCA()
    {
        #ifdef EHS_ARCH_X64
            return GetFeatureBits_2() & 0b00000000000001000000000000000000;
        #else
	        return false;
        #endif
    }

    bool CPU::RetrieveSSE4_1()
    {
        #ifdef EHS_ARCH_X64
            return GetFeatureBits_2() & 0b00000000000010000000000000000000;
        #else
	        return false;
        #endif
    }

    bool CPU::RetrieveSSE4_2()
    {
        #ifdef EHS_ARCH_X64
            return GetFeatureBits_2() & 0b00000000000100000000000000000000;
        #else
	        return false;
        #endif
    }

    bool CPU::RetrieveX2APIC()
    {
        #ifdef EHS_ARCH_X64
            return GetFeatureBits_2() & 0b00000000001000000000000000000000;
        #else
	        return false;
        #endif
    }

    bool CPU::RetrieveMOVBE()
    {
        #ifdef EHS_ARCH_X64
            return GetFeatureBits_2() & 0b00000000010000000000000000000000;
        #else
	        return false;
        #endif
    }

    bool CPU::RetrievePOPCNT()
    {
        #ifdef EHS_ARCH_X64
            return GetFeatureBits_2() & 0b00000000100000000000000000000000;
        #else
	        return false;
        #endif
    }

    bool CPU::RetrieveTSC_DEADLINE()
    {
        #ifdef EHS_ARCH_X64
            return GetFeatureBits_2() & 0b00000001000000000000000000000000;
        #else
	        return false;
        #endif
    }

    bool CPU::RetrieveAES()
    {
        #ifdef EHS_ARCH_X64
            return GetFeatureBits_2() & 0b00000010000000000000000000000000;
        #else
	        return false;
        #endif
    }

    bool CPU::RetrieveXSAVE()
    {
        #ifdef EHS_ARCH_X64
            return GetFeatureBits_2() & 0b00000100000000000000000000000000;
        #else
	        return false;
        #endif
    }

    bool CPU::RetrieveOSXSAVE()
    {
        #ifdef EHS_ARCH_X64
            return GetFeatureBits_2() & 0b00001000000000000000000000000000;
        #else
	        return false;
        #endif
    }

    bool CPU::RetrieveAVX()
    {
        #ifdef EHS_ARCH_X64
            return GetFeatureBits_2() & 0b00010000000000000000000000000000;
        #else
	        return false;
        #endif
    }

    bool CPU::RetrieveF16C()
    {
        #ifdef EHS_ARCH_X64
            return GetFeatureBits_2() & 0b00100000000000000000000000000000;
        #else
	        return false;
        #endif
    }

    bool CPU::RetrieveRDRND()
    {
        #ifdef EHS_ARCH_X64
            return GetFeatureBits_2() & 0b01000000000000000000000000000000;
        #else
	        return false;
        #endif
    }

    bool CPU::RetrieveHYPERVISOR()
    {
        #ifdef EHS_ARCH_X64
            return GetFeatureBits_2() & 0b10000000000000000000000000000000;
        #else
	        return false;
        #endif
    }

    bool CPU::RetrieveAVX2()
    {
        #ifdef EHS_ARCH_X64
            return GetExtFeatureBits_1() & 0b00000000000000000000000000100000;
        #else
	        return false;
        #endif
    }

    bool CPU::RetrieveAVX512F()
    {
        #ifdef EHS_ARCH_X64
	            return GetExtFeatureBits_1() & 0b00000000000000010000000000000000;
        #else
	            return false;
        #endif
    }

    bool CPU::RetrieveRDSEED()
    {
        #ifdef EHS_ARCH_X64
            return GetExtFeatureBits_1() & 0b00000000000001000000000000000000;
        #else
	        return false;
        #endif
    }

    bool CPU::RetrieveADX()
    {
        #ifdef EHS_ARCH_X64
            return GetExtFeatureBits_1() & 0b00000000000010000000000000000000;
	    #else
	        return false;
	    #endif
    }

	/*
	Str_8 CPU::ToStr()
	{
		return "Manufacturer: " + GetManufacturer() + "\r\n" +
				"Brand: " + GetBrand() + "\r\n" +
				"Stepping Id: " + Str_8::FromNum(GetSteppingId()) + "\r\n" +
				"GpuModel Id: " + Str_8::FromNum(GetModelId()) + "\r\n" +
				"Family Id: " + Str_8::FromNum(GetFamilyId()) + "\r\n" +
				"Processor Type Id: " + Str_8::FromNum(GetProcessorTypeId()) + "\r\n" +
				"Extended GpuModel Id: " + Str_8::FromNum(GetExtModelId()) + "\r\n" +
				"Extended Family Id: " + Str_8::FromNum(GetExtFamilyId()) + "\r\n" +
				"Has FPU: " + Str_8::FromNum((UInt_8)HasFPU()) + "\r\n" +
				"Has SSE: " + Str_8::FromNum((UInt_8)HasSSE()) + "\r\n" +
				"Has SSE 2: " + Str_8::FromNum((UInt_8)HasSSE2()) + "\r\n" +
				"Has SSE 3: " + Str_8::FromNum((UInt_8)HasSSE3()) + "\r\n" +
				"Has SSSE 3: " + Str_8::FromNum((UInt_8)HasSSSE3()) + "\r\n" +
				"Has SSE 4.1: " + Str_8::FromNum((UInt_8)HasSSE4_1()) + "\r\n" +
				"Has SSE 4.2: " + Str_8::FromNum((UInt_8)HasSSE4_2()) + "\r\n" +
				"Has AVX: " + Str_8::FromNum((UInt_8)HasAVX()) + "\r\n" +
				"Has RDRND: " + Str_8::FromNum((UInt_8)HasRDRND()) + "\r\n" +
				"Has AVX 2: " + Str_8::FromNum((UInt_8)HasAVX2()) + "\r\n" +
				"Has ADX: " + Str_8::FromNum((UInt_8)HasADX()) + "\r\n" +
				"Has RDSEED: " + Str_8::FromNum((UInt_8)HasRDSEED());
	}
	*/

	UInt_64 CPU::RetrieveTSC_Freq()
    {
        File tscDatabase("TSC_Frequencies.json", Mode::READ_WRITE, Disposition::CREATE_PERSISTENT);
        if (tscDatabase.Size())
        {
            Json json(tscDatabase.ReadStr_8(tscDatabase.Size()), 5);
            JsonObj* root = (JsonObj*)json.GetValue();

            Char_8 manu[13];
            manu[12] = 0;
            GetManufacturer(manu);

            JsonVar* jManu = root->GetVar(manu);
            if (jManu)
            {
                JsonObj* joManu = (JsonObj*)jManu->GetValue();

                Char_8 brand[48];
                GetBrand(brand);

                JsonVar* jBrand = joManu->GetVar(brand);
                if (jBrand)
                {
                    tscDatabase.Release();

                    return (UInt_64)*(JsonNum*)jBrand->GetValue();
                }
                else
                {
					UInt_64 tscFreq = CalculateTSC_Freq();

                    joManu->AddVar({brand, tscFreq});

                    tscDatabase.SeekBeginning();
                    tscDatabase.WriteStr_8(json.ToStr(false));
                    tscDatabase.Release();

                    return tscFreq;
                }
            }
            else
            {
				UInt_64 tscFreq = CalculateTSC_Freq();

                Char_8 brand[48];
                GetBrand(brand);

                JsonObj cpus(1, 0);
                cpus[0] = JsonVar(brand, tscFreq);

                JsonVar tmp({brand, cpus});
                root->AddVar(tmp);

                tscDatabase.SeekBeginning();
                tscDatabase.WriteStr_8(json.ToStr(false));
                tscDatabase.Release();

                return tscFreq;
            }
        }

		UInt_64 tscFreq = CalculateTSC_Freq();

        Char_8 manu[13];
        manu[12] = 0;
        GetManufacturer(manu);

        Char_8 brand[49];
        brand[48] = 0;
        GetBrand(brand);

        JsonObj jManu(1, 0);
        jManu[0] = JsonVar(brand, tscFreq);

        JsonObj root(1, 0);
        root[0] = JsonVar(manu, jManu);

        Json json(root);

        tscDatabase.WriteStr_8(json.ToStr(false));
        tscDatabase.Release();

        return tscFreq;
    }

	UInt_64 CPU::CalculateTSC_Freq()
    {
		UInt_64 result = GetTSC();

        Thread::SleepFor(10000);

        return (GetTSC() - result) / 10;
    }
}